/* powmod-x86-64.S -- (C) Geoffrey Reynolds, February 2007.

   uint64_t powmod64_x86_64(uint64_t b, uint64_t n, uint64_t p, double invp);

     Returns b^n (mod p), where 0 <= b < p < 2^51.

     Assumes invp = 1.0/p (computed in round-to-zero mode).
     Assumes current SSE rounding mode is round-to-zero.


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include "config.h"

#if NEED_UNDERSCORE
#define powmod64_x86_64 _powmod64_x86_64
#endif

#ifdef _WIN64
#define ARG1 %rcx
#define ARG2 %rdx
#define ARG2l %edx
#define ARG2b %dl
#define ARG3 %r8
#define ARG4 %xmm3
#define REG5 %r9
#else
#define ARG1 %rdi
#define ARG2 %rsi
#define ARG2l %esi
#define ARG2b %sil
#define ARG3 %rdx
#define ARG4 %xmm0
#define REG5 %rcx
#endif

	.text
	.p2align 4,,15
	.globl powmod64_x86_64

powmod64_x86_64:
	mov	$1, %eax
	cmp	%rax, ARG2
	jbe	skip
	cvtsi2sdq ARG1, %xmm1
	test	%eax, ARG2l
	cmovnz	ARG1, %rax
	jmp	sqr
skip:
	cmovz	ARG1, %rax
	ret

	/* %rax = a, ARG1 = b, ARG2 = n, ARG3 = p, ARG4 = 1.0/p */
	.p2align 4,,15
loop:
	testb	$1, ARG2b
	cvtsi2sdq ARG1, %xmm1
	jz	sqr
mul:
	cvtsi2sdq %rax, %xmm2
	imul	ARG1, %rax
	mulsd	%xmm1, %xmm2
	mulsd	ARG4, %xmm2
	cvtsd2siq %xmm2, REG5
	imul	ARG3, REG5
	sub	REG5, %rax
	mov	%rax, REG5
	sub     ARG3, REG5
#if USE_CMOV
	cmovge	REG5, %rax
#else
	jl      sqr
	mov     REG5, %rax
#endif
sqr:
	imul	ARG1, ARG1
	mulsd	%xmm1, %xmm1
	mulsd	ARG4, %xmm1
	cvtsd2siq %xmm1, REG5
	shr	ARG2
	imul	ARG3, REG5
	sub	REG5, ARG1
	mov	ARG1, REG5
	sub     ARG3, REG5
#if USE_CMOV
	cmovge	REG5, ARG1
#else
	jl	test
	mov     REG5, ARG1
#endif
test:
	test	ARG2, ARG2
	jnz	loop

	ret
