/* powmod-x87-64.S -- (C) Geoffrey Reynolds, February 2007.

   uint64_t powmod64_x87_64(uint64_t b, uint64_t n, uint64_t p);

     Returns b^n (mod p), where 0 <= a < p < 2^62.

     Assumes FPU is set to double extended precision and round to zero.
     Assumes %st(0) contains 1.0/p computed with above settings.


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include "config.h"

#if NEED_UNDERSCORE
#define powmod64_x87_64 _powmod64_x87_64
#endif

#ifdef _WIN64
#define ARG1 %rcx
#define ARG2 %rdx
#define ARG2l %edx
#define ARG2b %dl
#define ARG3 %r8
#define REG4 %r9
/* Shadow space */
#define VAR1 8(%rsp)
#define VAR2 16(%rsp)
#else
#define ARG1 %rdi
#define ARG2 %rsi
#define ARG2l %esi
#define ARG2b %sil
#define ARG3 %rdx
#define REG4 %rcx
/* Red zone */
#define VAR1 -8(%rsp)
#define VAR2 -16(%rsp)
#endif

	.text
	.p2align 4,,15
	.globl powmod64_x87_64

powmod64_x87_64:
	mov	$1, %eax
	cmp	%rax, ARG2
	jbe	skip
	mov	ARG1, VAR1
	test	%eax, ARG2l
	fildll  VAR1
	cmovnz	ARG1, %rax
	jmp	sqr
skip:
	cmovz	ARG1, %rax
	ret

	/* %rax = a, ARG1 = b, ARG2 = n, ARG3 = p, %st(0) = 1.0/p */
	.p2align 4,,15
loop:
	mov	ARG1, VAR1
	mov	%rax, VAR2
	test	$1, ARG2b
	fildll  VAR1
	jz	sqr
mul:
	fildll  VAR2
	fmul    %st(1), %st(0)
	fmul    %st(2), %st(0)
	fistpll VAR2
	imul    ARG1, %rax
	mov     VAR2, REG4
	imul    ARG3, REG4
	sub     REG4, %rax
	mov     %rax, REG4
	sub     ARG3, REG4
#if USE_CMOV
	cmovge	REG4, %rax
#else
	jl      sqr
	mov     REG4, %rax
#endif
sqr:
	fmul    %st(0), %st(0)
	fmul    %st(1), %st(0)
	fistpll VAR2
	imul    ARG1, ARG1
	shr	ARG2
	mov     VAR2, REG4
	imul    ARG3, REG4
	sub     REG4, ARG1
	mov     ARG1, REG4
	sub     ARG3, REG4
#if USE_CMOV
	cmovge	REG4, ARG1
#else
	jl	test
	mov     REG4, ARG1
#endif
test:
	test	ARG2, ARG2
	jnz	loop

	ret
